from __future__ import annotations

import asyncio
import base64
import copy
import hashlib
import json
import mimetypes
import os
import pkgutil
import secrets
import shutil
import tempfile
import warnings
from concurrent.futures import CancelledError
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from threading import Lock
from typing import Any, Callable, Literal, Optional, TypedDict

import fsspec.asyn
import httpx
import huggingface_hub
from huggingface_hub import SpaceStage
from websockets.legacy.protocol import WebSocketCommonProtocol

class TooManyRequestsError(Exception):
    """Raised when the API returns a 429 status code."""

    pass


class QueueError(Exception):
    """Raised when the queue is full or there is an issue adding a job to the queue."""

    pass


class InvalidAPIEndpointError(Exception):
    """Raised when the API endpoint is invalid."""

    pass


class SpaceDuplicationError(Exception):
    """Raised when something goes wrong with a Space Duplication."""

    pass


class ServerMessage(str, Enum):
    send_hash = "send_hash"
    queue_full = "queue_full"
    estimation = "estimation"
    send_data = "send_data"
    process_starts = "process_starts"
    process_generating = "process_generating"
    process_completed = "process_completed"
    log = "log"
    progress = "progress"
    heartbeat = "heartbeat"
    server_stopped = "server_stopped"
    unexpected_error = "unexpected_error"


class Status(Enum):
    """Status codes presented to client users."""

    STARTING = "STARTING"
    JOINING_QUEUE = "JOINING_QUEUE"
    QUEUE_FULL = "QUEUE_FULL"
    IN_QUEUE = "IN_QUEUE"
    SENDING_DATA = "SENDING_DATA"
    PROCESSING = "PROCESSING"
    ITERATING = "ITERATING"
    PROGRESS = "PROGRESS"
    FINISHED = "FINISHED"
    CANCELLED = "CANCELLED"
    LOG = "LOG"

    @staticmethod
    def ordering(status: Status) -> int:
        """Order of messages. Helpful for testing."""
        order = [
            Status.STARTING,
            Status.JOINING_QUEUE,
            Status.QUEUE_FULL,
            Status.IN_QUEUE,
            Status.SENDING_DATA,
            Status.PROCESSING,
            Status.PROGRESS,
            Status.ITERATING,
            Status.FINISHED,
            Status.CANCELLED,
        ]
        return order.index(status)

    def __lt__(self, other: Status):
        return self.ordering(self) < self.ordering(other)

    @staticmethod
    def msg_to_status(msg: str) -> Status:
        """Map the raw message from the backend to the status code presented to users."""
        return {
            ServerMessage.send_hash: Status.JOINING_QUEUE,
            ServerMessage.queue_full: Status.QUEUE_FULL,
            ServerMessage.estimation: Status.IN_QUEUE,
            ServerMessage.send_data: Status.SENDING_DATA,
            ServerMessage.process_starts: Status.PROCESSING,
            ServerMessage.process_generating: Status.ITERATING,
            ServerMessage.process_completed: Status.FINISHED,
            ServerMessage.progress: Status.PROGRESS,
            ServerMessage.log: Status.LOG,
            ServerMessage.server_stopped: Status.FINISHED,
        }[msg]  # type: ignore


@dataclass
class ProgressUnit:
    index: Optional[int]
    length: Optional[int]
    unit: Optional[str]
    progress: Optional[float]
    desc: Optional[str]

    @classmethod
    def from_msg(cls, data: list[dict]) -> list[ProgressUnit]:
        return [
            cls(
                index=d.get("index"),
                length=d.get("length"),
                unit=d.get("unit"),
                progress=d.get("progress"),
                desc=d.get("desc"),
            )
            for d in data
        ]


@dataclass
class StatusUpdate:
    """Update message sent from the worker thread to the Job on the main thread."""

    code: Status
    rank: int | None
    queue_size: int | None
    eta: float | None
    success: bool | None
    time: datetime | None
    progress_data: list[ProgressUnit] | None
    log: tuple[str, str] | None = None


def create_initial_status_update():
    return StatusUpdate(
        code=Status.STARTING,
        rank=None,
        queue_size=None,
        eta=None,
        success=None,
        time=datetime.now(),
        progress_data=None,
    )


@dataclass
class JobStatus:
    """The job status.

    Keeps track of the latest status update and intermediate outputs (not yet implements).
    """

    latest_status: StatusUpdate = field(
        default_factory=create_initial_status_update)
    outputs: list[Any] = field(default_factory=list)


@dataclass
class Communicator:
    """Helper class to help communicate between the worker thread and main thread."""

    lock: Lock
    job: JobStatus
    prediction_processor: Callable[..., tuple]
    reset_url: str
    should_cancel: bool = False
    event_id: str | None = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import concurrent.futures
import json
import os
import re
import secrets
import tempfile
import threading
import time
import urllib.parse
import uuid
import warnings
from concurrent.futures import Future
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from threading import Lock
from typing import Any, Callable
from .queue_client import QueueClient
from .utils import (
    Communicator,
    JobStatus,
    Message,
    QueueError,
    ServerMessage,
    Status,
    StatusUpdate,
)


class Client:
    def __init__(self, endpoint, token, input_name, max_workers: int = 40):
        self.endpoint = endpoint
        self.token = token
        self.input_name = input_name
        self.sink_name = input_name + '/sink'
        self.input_queue = QueueClient(endpoint, input_name)
        self.sink_queue = QueueClient(endpoint, self.sink_name)
        self.input_queue.set_token(token)
        self.input_queue.init()
        self.sink_queue.set_token(token)
        self.sink_queue.init()
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=max_workers)
        
    def make_predict(self, helper: Communicator | None = None):
        def _predict(data, tags, req_path):
            index, request_id = self.input_queue.put(data, tags=tags)
            # print(f'input index: {index}, data: {data}')
            result = None
            while result is None:
                try:
                    dfs = self.sink_queue.get(request_id=request_id, timeout='5s', auto_delete=False)
                    if len(dfs) > 0:
                        result = str(dfs[0].data)
                        self.sink_queue.delete(dfs[0].index)
                        # print(f'sink index: {dfs[0].index}, data: {dfs[0].data}')
                        return result
                except Exception as e:
                    print("err:", e)
                    continue

        return _predict

    def submit(self, data, tags: dict = {}, req_path = '', result_callbacks=None):
        predict_fn = self.make_predict()
        future = self.executor.submit(predict_fn, data, tags, req_path)

        job = Job(future, communicator=None)
        if result_callbacks:
            if isinstance(result_callbacks, Callable):
                result_callbacks = [result_callbacks]

            def create_fn(callback) -> Callable:
                def fn(future):
                    if isinstance(future.result(), tuple):
                        callback(*future.result())
                    else:
                        callback(future.result())

                return fn

            for callback in result_callbacks:
                job.add_done_callback(create_fn(callback))

        return job


class Job(Future):

    def __init__(
        self,
        future: Future,
        communicator: Communicator | None = None
    ):
        self.future = future
        self.communicator = communicator

    def result(self, timeout: float | None = None) -> Any:
        return super().result(timeout=timeout)

    def status(self) -> StatusUpdate:
        time = datetime.now()
        cancelled = False
        if self.communicator:
            with self.communicator.lock:
                cancelled = self.communicator.should_cancel
        if cancelled:
            return StatusUpdate(
                code=Status.CANCELLED,
                rank=0,
                queue_size=None,
                success=False,
                time=time,
                eta=None,
                progress_data=None,
            )
        if self.done():
            if not self.future._exception:  # type: ignore
                return StatusUpdate(
                    code=Status.FINISHED,
                    rank=0,
                    queue_size=None,
                    success=True,
                    time=time,
                    eta=None,
                    progress_data=None,
                )
            else:
                return StatusUpdate(
                    code=Status.FINISHED,
                    rank=0,
                    queue_size=None,
                    success=False,
                    time=time,
                    eta=None,
                    progress_data=None,
                )
        elif not self.communicator:
            return StatusUpdate(
                code=Status.PROCESSING,
                rank=0,
                queue_size=None,
                success=None,
                time=time,
                eta=None,
                progress_data=None,
            )
        else:
            with self.communicator.lock:
                return self.communicator.job.latest_status

    def cancel(self) -> bool:
        if self.communicator:
            with self.communicator.lock:
                self.communicator.should_cancel = True
                return True
        return self.future.cancel()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from concurrent.futures import ThreadPoolExecutor
import threading
import time
from .queue_client import QueueClient

class AsyncResult:
    def __init__(self):
        self._result = None
        self._callbacks = []
        self._is_finished = threading.Event()

    def set_result(self, result):
        self._result = result
        self._is_finished.set()
        for callback in self._callbacks:
            callback(result)

    def add_done_callback(self, callback):
        if self._is_finished.is_set():
            callback(self._result)
        else:
            self._callbacks.append(callback)

    def wait(self, timeout=None):
        self._is_finished.wait(timeout)
        return self._result

class Client:
    def __init__(self, endpoint, token, input_name):
        self.endpoint = endpoint
        self.token = token
        self.input_name = input_name
        self.sink_name = input_name + '/sink'
        self.input_queue = QueueClient(endpoint, input_name)
        self.sink_queue = QueueClient(endpoint, self.sink_name)
        self.input_queue.set_token(token)
        self.input_queue.init()
        self.sink_queue.set_token(token)
        self.sink_queue.init()
        self.executor = ThreadPoolExecutor()

    def async_predict(self, data, callback=None):
        index, request_id = self.input_queue.put(data)
        async_result = AsyncResult()
        if callback:
            async_result.add_done_callback(callback)

        # The asynchronous task that will be submitted to the executor
        def task():
            result = None
            while result is None:
                dfs = self.sink_queue.get(request_id=request_id, timeout='0s', auto_delete=False)
                if len(dfs) > 0:
                    result = str(dfs[0].data)
                    self.sink_queue.delete(dfs[0].index)
                else:
                    time.sleep(1)  # Polling interval
            async_result.set_result(result)

        self.executor.submit(task)
        return async_result
